import numpy as np
from copy import deepcopy


class Board:
    blocked = 0
    allowed = 1
    queen = 2
    
    def __init__(self, x: int, y: int):
        self.board = np.ones((x, y), dtype=int)
        self.dimX = x
        self.dimY = y
        
    def copy(self):
        return deepcopy(self)
    
    def toConsole(self) -> None:
        for y in range(self.dimY):
            line = ""
            for x in range(self.dimX):
                if self.board[x, y] == self.blocked:
                    line += "X"
                if self.board[x, y] == self.allowed:
                    line += "O"
                if self.board[x, y] == self.queen:
                    line += "@"
            print(line)
        print("")
    
    def print(self) -> None:
        self.toConsole()
    
    def get_allowed_fields(self):
        allowed_fields = []
        for x in range(self.dimX):
            for y in range(self.dimY):
                if self.board[x, y] == self.allowed:
                    allowed_fields.append((x, y))
        return allowed_fields
    
    def get_queen_fields(self):
        queen_fields = []
        for x in range(self.dimX):
            for y in range(self.dimY):
                if self.board[x, y] == self.allowed:
                    queen_fields.append((x, y))
        return queen_fields
    
    def put_queen(self, x: int, y: int) -> bool:
        if self.board[x, y] == self.allowed:
            for _x in range(self.dimX):
                self.board[_x, y] = self.blocked
            for _y in range(self.dimY):
                self.board[x, _y] = self.blocked

            _x = x
            _y = y
            while _x > 0 and _y > 0:
                _y -= 1
                _x -= 1
                self.board[_x, _y] = self.blocked
            
            _x = x
            _y = y
            while _x < self.dimX-1 and _y < self.dimY-1:
                _y += 1
                _x += 1
                self.board[_x, _y] = self.blocked
            self.board[x, y] = self.queen
            return True
        else:
            return False


if __name__ == "__main__":
    board = Board(8, 8)
    
    saved_boards = []
    iters = 0
    while board.get_allowed_fields().__len__() > 0:
        board.put_queen(board.get_allowed_fields()[0][0],
                        board.get_allowed_fields()[0][1])
        iters += 1
        board.toConsole()
        saved_boards.append(board.copy())
    print("Queens: " + str(iters))
