import matplotlib.pyplot as plt
from math import *
from random import randint
from string import ascii_uppercase
from itertools import permutations
# config:
brute_force = False
optimize = True
num_nodes = 10
do_plot = False
pause = 1

class Node:
    x: int
    y: int
    name: str
    def __init__(self, x: int, y: int, name: str = ""):
        self.x = x
        self.y = y
        self.name = name
    
    def plot(this):
        plt.plot(this.x, this.y, 'ro', label=this.name)

def distance(A: Node, B: Node) -> float:
    return sqrt((A.x - B.x)**2 + (A.y - B.y)**2)

def total_distance(route: list) -> float:
    total_distance = 0
    for i in range(len(route)):
        if i == len(route)-1:
            total_distance += distance(route[i], route[0])
        else:
            total_distance += distance(route[i], route[i+1])
    return total_distance

class Path:
    A: Node
    B: Node
    distance: float
    def __init__(self, A: Node, B: Node):
        self.A = A
        self.B = B
        #self.distance = distance(A, B)
    
    def plot(this):
        plt.plot([this.A.x, this.B.x], [this.A.y, this.B.y], 'go-')


def plot_route(route: list, nodes: list = [], circle: bool = False) -> None:
    plt.close("all")
    for i in range(len(route)):
        if i == len(route) - 1 and circle:
            Path(A=route[i], B=route[0]).plot()
        else:
            Path(A=route[i], B=route[i+1]).plot()
    for node in nodes:
        node.plot()
    plt.show()
    plt.pause(pause)


def brute(nodes: list) -> list:
    if len(nodes) <= 1:
        return
    best_route = nodes
    best_route_length = total_distance(nodes)
    if do_plot:
        plot_route(best_route)
#TODO Noch nicht rekursiv!
    for i, route in enumerate(permutations(nodes, len(nodes))):
        route_length = total_distance(route)
        if route_length < best_route_length:
            best_route = route
            best_route_length = route_length
            print("Found shorter Route with length " + str(best_route_length) + " after " +  str(i) + " iterations.")
            if do_plot:
                plot_route(best_route)
    print(str(i+1) + " iterations in total")
    return best_route


def get_closest_node(current_node: Node, nodes: list) -> Node:
    if len(nodes) == 0:
        return
    closest_node = nodes[0]
    closest_distance = distance(current_node, nodes[0])
    for node in nodes:
        dist = distance(current_node, node)
        if dist < closest_distance:
            closest_node = node
            closest_distance = dist
    return closest_node

def nearest_neighbour(_nodes: list) -> list:
    remaining_nodes = _nodes.copy()
    route = []
    if len(remaining_nodes) == 0:
        return route
    route.append(remaining_nodes[0])
    last_node = remaining_nodes[0]
    remaining_nodes.remove(last_node)
    while(len(remaining_nodes) > 0):
        next_node = get_closest_node(last_node, remaining_nodes)
        remaining_nodes.remove(next_node)
        route.append(next_node)
        last_node = next_node
        if do_plot:
            plot_route(route, _nodes)
    return route


def generate_paths(nodes: list) -> list:
    _nodes = nodes.copy()
    paths = []
    if _nodes.__len__() <= 1:
        return []

    this_node = _nodes[0]
    _nodes.remove(_nodes[0])
    
    for node in _nodes:
        paths.append(Path(A=this_node, B=node))
    
    if _nodes.__len__() >= 2:
        paths = paths + generate_paths(_nodes)
    
    return paths


#TODO Zuweisungen brechen die Kette!
def optimize_2opt(_route: list) -> list:
    route = _route.copy()
    
    for i in range(len(route)-1):
        a = route[i]
        b = route[i+1]
        ab  = distance(a, b)
        
        for j in range(2, len(route)-1):
            swapped = False
            c = route[j]
            d = route[j+1]
            
            cd = distance(c, d)
            abcd = ab + cd
            
            ac = distance(a, c)
            db = distance(d, b)
            acdb = ac + db
            
            ad = distance(a, d)
            cb = distance(c, b)
            adcb = ad + cb
            
            if adcb < acdb:
                if adcb < abcd:
                    #adcb
                    #swap d & b
                    route[i+1] = d
                    route[j+1] = b
                    #abcd
                    swapped = True
            else:
                if acdb < abcd:
                    #acdb
                    #swap c & b
                    route[j] = b
                    route[i+1] = c
                    #abdc
                    #swap c & d
                    route[j] = d
                    route[j+1]  = c
                    #abcd
                    swapped = True
            if swapped:
                break

    
    return route



if __name__ == "__main__":
    nodes = []
    for i in range(num_nodes):
        nodes.append(Node(x=randint(0, 100), y=randint(0, 100), name=str(hex(i))))
        

    plt.ion()
    
    if brute_force:
        best_route = brute(nodes)
    else:
        best_route = nearest_neighbour(nodes)
        if optimize:
            best_route = optimize_2opt(best_route)
    
    
    plot_route(best_route, nodes, circle=True)
    plt.show(block=True)
